/*
**  Copyright (c) 2003,2004,2005 Andrea Luzzardi <scox@gentoo.org>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <shadow.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utmp.h>

#define PAM_SM_AUTH

#include <security/pam_modules.h>
#include <security/_pam_macros.h>

#include "pam.h"
#include "parser.h"
#include "conf.h"
#include "device.h"
#include "auth.h"

char *converse (pam_handle_t *pamh, int style, char *message)
{
	struct pam_conv *c;
	struct pam_message pam_msg;
	struct pam_response *reply = NULL;

	const struct pam_message *msg_ptr;

	union {
		struct pam_conv *c;
	} cnv;

	if (pam_get_item (pamh, PAM_CONV, (const void **)&cnv.c) != PAM_SUCCESS) {
		DEBUGP ("Cannot start conversation\n");
		return NULL;
	}

	pam_msg.msg_style = style;
	pam_msg.msg = message;
	msg_ptr = &pam_msg;

	c = cnv.c;

	c->conv (1, &msg_ptr, &reply, c->appdata_ptr);

	if (reply && style != PAM_PROMPT_ECHO_OFF)
		_pam_drop_reply (reply, 1);

	return reply->resp;
}

int local_login (const char *login, pam_handle_t *pamh)
{
	struct utmp *utent = NULL;
	char **local = NULL;
	char *tty = NULL;
	int ret = 1, i = 0;

	if (usb_conf.allow_remote == 1)
		return 1;

	tty = ttyname(fileno(stdin));

	if (!tty) {
		INFOP("Authentication denied: Cannot retrieve tty name, rejecting.\n");
		return 0;
	}

	local = split (usb_conf.local_consoles, ',');

	while (local[i]) {
		if (!strncmp (tty, local[i], strlen(local[i]))) {
			INFOP ("Authentication request for user %s from local console" \
					" (%s)\n", login, tty);
			drop_split (local);
			return 1;
		}

		++i;
	}

	if (!strncmp(tty, "/dev/", strlen("/dev/")))
		tty += strlen("/dev/");

	i = 0;
	drop_split (local);

	utmpname (usb_conf.utmp);
	setutent();

	DEBUGP ("Searching the utmp entry for tty %s...\n", tty);

	while ((utent = getutent ()) != NULL) {
		if (!(ret = strcmp (tty, utent->ut_line)))
			break;
	}

	endutent();

	if (ret != 0) {
		DEBUGP ("Cannot retrieve the utmp entry\n");
		return 0;
	}

	INFOP ("Authentication request for user %s (%s %s)\n", login,
			tty, utent->ut_host);

	if (!strlen(utent->ut_host))
		return 1;

	local = split (usb_conf.local_hosts, ',');

	while (local[i]) {
		if (!strncmp (utent->ut_host, local[i], strlen(local[i]))) {
			drop_split (local);
			return 1;
		}

		++i;
	}

	drop_split (local);

	return 0;
}

PAM_EXTERN
int pam_sm_authenticate (pam_handle_t *pamh, int flags, int argc,
		const char **argv)
{
	const char *user = NULL;

	if (pam_get_user (pamh, &user, NULL) != PAM_SUCCESS)
		return PAM_AUTH_ERR;

	/* Do we trust data returned by PAM functions ? */

	if (!user || strlen(user) <= 0)
		return PAM_AUTH_ERR;

	parse_args (argc, argv);
	put_default_values();

	log_init();

	INFOP("pam_usb v%s, (C) 2003-2005 Andrea Luzzardi <scox@sig11.org>\n",
			PVERSION);

	if (!(local_login(user, pamh))) {
		INFOP ("Authentication denied: remote user.\n");
		return PAM_AUTH_ERR;
	}

	if (!(check_device())) {
		INFOP("Authentication denied: Invalid device.\n");
		return PAM_AUTHINFO_UNAVAIL;
	}

	if (!authenticate(pamh, user)) {
		DEBUGP ("Cannot authenticate user \"%s\"\n", user);
		return PAM_AUTH_ERR;
	}

	INFOP("Access granted\n\n");

	log_fini();
	
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_setcred (pam_handle_t *pamh,int flags,int argc, const char **argv)
{
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_acct_mgmt (pam_handle_t *pamh, int flags, int argc,
			const char **argv)
{
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_chauthtok (pam_handle_t *pamh, int flags, int argc,
			const char **argv)
{
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_open_session (pam_handle_t *pamh, int flags, int argc,
			const char **argv)
{
	return PAM_SUCCESS;
}

PAM_EXTERN
int pam_sm_close_session (pam_handle_t *pamh, int flags, int argc,
			const char **argv)
{
	return PAM_SUCCESS;
}

#ifdef PAM_STATIC

struct pam_module _pam_usb_modstruct = {
	"pam_usb",
	pam_sm_authenticate,
	pam_sm_setcred,
	pam_sm_acct_mgmt,
	pam_sm_open_session,
	pam_sm_close_session,
	pam_sm_chauthtok
};

#endif
