/*
**  Copyright (c) 2003,2004,2005 Andrea Luzzardi <scox@gentoo.org>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "usbadm.h"
#include <shadow.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <pwd.h>
#include <unistd.h>
#include <security/pam_modules.h>
#include <security/_pam_macros.h>

#include <device.h>
#include <parser.h>
#include <dsa.h>

struct passwd *get_usr_nfo (char *user)
{
	struct passwd *entry = NULL;

	while ((entry = getpwent ()))
		if (!strcmp (user, entry->pw_name) &&
		    !strcmp (entry->pw_name, user))
			break;
	setpwent();
	return entry;
}

int dir_check (char *dir, char *keypath, uid_t owner, gid_t group)
{
	char *auth_dir = NULL;
	struct stat buf;

	auth_dir = insert_before (dir, keypath);
	if (!auth_dir)
		return 0;

	if (stat (auth_dir, &buf) == 0) {
		free (auth_dir);
		return 1;
	}

	printf ("[!] Directory %s not found, creating one...\n", auth_dir);

	if (mkdir (auth_dir, S_IRUSR | S_IWUSR | S_IXUSR) == -1) {
		free (auth_dir);
		return 0;
	}

	if (chown (auth_dir, owner, group) == -1) {
		printf ("[W] Cannot chown u(%d) g(%d) the directory %s\n",
			owner, group, auth_dir);
		free (auth_dir);
		return 1;
	}

	free (auth_dir);
	return 1;
}

int check_tree (char *user, char *mnt)
{
	struct passwd *entry = NULL;

	if (!(entry = get_usr_nfo(user)))
		return 0;

	if (!dir_check (entry->pw_dir, usb_conf.local_keypath, entry->pw_uid,
					entry->pw_gid))
		return 0;

	if (!dir_check (mnt, usb_conf.device_keypath, 0, 0))
		return 0;

	return 1;
}

int gen_dsa_keys (int bits, FILE *priv, FILE *pub)
{
	DSA *key;

	key = DSA_generate_parameters (bits, NULL, 0, NULL, NULL, NULL, NULL);

	if (!key) {
		printf ("[-] Failed to generate private key's parameters\n");
		return 0;
	}

	printf ("[!] Extracting private key...\n");

	if (!DSA_generate_key (key)) {
		printf ("[-] Failed to extract private key\n");
		return 0;
	}

	printf ("[+] Private key extracted.\n");

	if (!PEM_write_DSAPrivateKey (priv, key, NULL, NULL, 0, NULL, NULL)) {
		printf ("[-] Cannot write DSA private key.\n");
		return 0;
	}

	printf ("[+] Private key successfully written.\n");

	printf ("[!] Writing public key...\n");

	if (!PEM_write_DSA_PUBKEY (pub, key)) {
		printf ("[-] Cannot write DSA public key.\n");
		return 0;
	}

	printf ("[+] Public key successfully written.\n");

	return 1;
}

int keygen (int ac, char **av)
{
	FILE *pub, *priv;

	char hostname[32];
	char *user, *file;
	int bits;

	if (ac < 5) {
		print_usage (av[1], av);
		return 1;
	}

	if (gethostname (hostname, 32) != 0) {
		printf ("[-] Cannot get hostname\n");
		return 1;
	}

	user = av[3];
	bits = atoi (av[4]);

	if (!(check_tree (user, av[2]))) {
		printf ("[-] Unable to locate and/or create keys directories\n");
		return 1;
	}

	file = get_private_filename (user, av[2]);

	if (!(priv = fopen (file, "w"))) {
		printf ("[-] Cannot open private keyfile \"%s\"\n", file);
		free (file);
		return 1;
	}

	free (file);

	file = get_public_filename (user);

	if (!(pub = fopen (file, "w"))) {
		printf ("[-] Cannot open public keyfile \"%s\"\n", file);
		free (file);
		fclose (priv);
		return 1;
	}

	free (file);

	printf ("[!] Generating %d DSA key pair for %s@%s\n", bits, user, hostname);
	if (!gen_dsa_keys (bits, priv, pub)) {
		fclose (priv);
		fclose (pub);
		return 1;
	}

	fclose (priv);
	fclose (pub);
	return 0;
}
